# TODO: if included in field_name as part and multiple values by 's', 'es' at the end
#   like {'field': 'languages', 'generator': '[choice(languages), choice(languages)]'}

default_rules = {'id': {'generator': 'uuid4().hex', 'len': 12},
                 'default': {'generator': None, 'len': 6},
                 'price': {'generator': 'round(random(), 2)'},
                 'ts': {'generator': 'current_time'},
                 'flag': {'generator': 'choice([\'Y\', \'N\'])'},
                 'date': {'generator': 'date.date(2019, 2021)'},
                 'city': {'generator': 'address.city()'},
                 'country': {'generator': 'address.country()'},
                 'language': {'generator': 'choice(languages)'},
                 'title': {'generator': 'text.text()', 'len': 24},
                 'name': {'generator': 'person.name().split()[0]'},
                 'description': {'generator': 'text.text()'},
                 'email': {'generator': 'person.email()'}
                 }



""" main module that contains RunGenerator class that used to init data generation """

from collections import defaultdict

from fakeme.config import Config
from fakeme.tables import MultiTableRunner
from fakeme.fields import FieldRulesExtractor

from datetime import datetime

target_paths = []
created = []
break_point = None


class RunGenerator(object):

    def __init__(self,
                 tables_list,
                 params=None,
                 dump_schema=True,
                 rls=None,
                 path_prefix=".",
                 appends=None,
                 cli_path=None):
        """
        Main Class to start data generation

        :param tables_list: list with tuples, what contain
            (dataset/database, table_id/table_name, path_to_table_schema
            or schema in one of supported formats), example:

                [
                    ("yourdatabase", "table_name1", "schemas/crm_paddr_hh.ddl"),
                    ("yourdatabase", "table_name2", schema_in_python_object),
                    ("yourdatabase_2", "table_name3", "schemas/your_schema_in_json.json") ...
                ]

                where
                        schema_in_python_object = [
                                {
                                  "type": "STRING",
                                  "name": "field1",
                                  "mode": "NULLABLE"
                                 },
                                 {
                                  "type": "STRING",
                                  "name": "field2"
                                 }]
        :type tables_list: list
        :param params:
        :type params: dict
        :param dump_schema:
        :type dump_schema:
        :param rls:
        :type rls:
        """
        print("Data Generation starts at", datetime.now())

        self.tables = tables_list
        self.cfg = Config(params).get_config()
        self.rls = self.resolve_aliases_tuples(rls) if rls else None or {}
        self.dump_schema = dump_schema
        self.path_prefix = path_prefix
        self.appends = appends or []
        self.cli_path = cli_path

    def resolve_aliases_tuples(self, rls: dict):
        """
            resolve fields aliases (situation when one table depend on second)
            or need to contain same value as in second table

            rls example:

                {'warehouse': {'part_id': {'alias': 'part_name',
                                                'matches': 1,
                                                'table': 'parts'}}}
        """

        for table in rls:
            new_dict = {}
            to_delete = []
            for key in rls[table]:
                if isinstance(key, tuple):
                    if isinstance(rls[table][key]["alias"], tuple):
                        for num, item in enumerate(key):
                            if len(rls[table][key]["alias"]) > num:
                                alias = rls[table][key]["alias"][num]
                            else:
                                alias = item
                            new_dict.update({item: {
                                "alias": alias,
                                "table": rls[table][key]["table"],
                                "matches": rls[table][key].get("matches")
                                            or self.cfg['matches']}})
                    to_delete.append(key)
            rls[table].update(new_dict)
            for item in to_delete:
                del rls[table][item]
        return rls

    def run(self):
        """
            call method to run data generation

            priority 0 - tables what target for chains
            priority 1 - tables what target for chains and in chains
            priority 2 - tables what not in self.append
            priority 3 - tables in self.append

        """

        print("Run Generation with settings: {}".format(self.cfg))
        # get all fields and schemas for tables
        self.schemas, fields = MultiTableRunner(
            self.tables, prefix=self.path_prefix, settings=self.cfg).\
            get_fields_and_schemas(dump_schema=self.dump_schema)

        priority_dict = self.create_priority_table_dict()
        field_extractor = FieldRulesExtractor(fields)

        # generate "value_rules.json" with rules for fields
        field_extractor.generate_rules()

        self.linked_fields = field_extractor.get_chains(self.schemas, self.rls)

        for level in priority_dict:
            for table in priority_dict[level]:
                if table not in created and table not in priority_dict.get(level+1, []):
                    self.create_table(table)
        if self.cfg['output'].get('line_start'):
            for target_file_path in target_paths:
                self._add_symbol_at_line_start(target_file_path)

        print("Data Generation ended successful \n", datetime.now())

    def _add_symbol_at_line_start(self, target_file_path):
        file_buf = []
        with open(target_file_path, "r") as csv_file:
            for num, line in enumerate(csv_file.readlines()):
                if num == 0:
                    if self.cfg["header"].get("no_quotes"):
                        line = line.replace('\"', "")
                    if self.cfg["header"].get("case"):
                        line = eval("line.{}()".format(self.cfg['header'].get(
                            "case")))
                file_buf.append(self.cfg['line_start'] + line)
        with open(target_file_path, "w+") as csv_file:
            csv_file.writelines(file_buf)

    def create_table(self, table):
        """ run table creation """
        target_path = "{}.{}".format(table, self.cfg['output']['file_format'])
        self.schemas[table][1].create_data(
            file_path=target_path, chained=self.linked_fields,
            alias_chain=self.rls, appends=self.appends,
            cli_path=self.cli_path)
        target_paths.append(target_path)
        created.append(table)

    def create_priority_table_dict(self):
        priority_dict = defaultdict(set)
        # 3
        [priority_dict[3].add(k) for k in self.appends]

        chains_tables = set([k for k in self.rls if k != "all"])

        priority_dict[0] = set([x[1] for x in self.tables if x[1] not in chains_tables])

        link_dict = {}
        tables_list = [x[1] for x in self.tables]
        for table in chains_tables:
            self_table = self.rls[table]
            for field in self_table:
                field_dict = self_table[field]
                if field_dict['table'] in tables_list:
                    if not link_dict.get(field_dict['table']):
                        link_dict[field_dict['table']] = set([])
                    link_dict[field_dict['table']].add(table)
        while link_dict:
            for n in range(0, 6):
                table_moved = []
                for table in link_dict:
                    if table in priority_dict[n]:
                        [priority_dict[n+1].add(tb_name) for tb_name in link_dict[table]]
                        table_moved.append(table)
                if table_moved:
                    for table in table_moved:
                        del link_dict[table]
        return priority_dict



"""  half of imports are used in step of field generation, so this is why imports looks like unused.

    TODO: need  to add lazy evaluation only if import required in generation rule field """
import os
import sys
import copy
import multiprocessing as mp
from multiprocessing import Queue
from datetime import datetime, timedelta
import pytz
from collections.abc import Iterable
import logging
from typing import Text, List, Dict
from pandas import DataFrame, Series, read_csv, read_json  # noqa F401

from fakeme.fields import FieldRules
from fakeme.values import values_generator

supported_types = ['STRING', 'INTEGER', 'FLOAT', 'LIST']

q = Queue()


class DataGenerator:
    def __init__(self,
                 schema: List,
                 settings: Dict = None,
                 chained: Dict = None,
                 table_id: Text = None,
                 alias_chain: Dict = None,
                 appends: Dict = None,
                 cli_path=None
                 ):
        self.schema = self.schema_validation(schema)
        self.settings = settings or {}
        self.chained = chained
        self.table_id = table_id
        self.chains = alias_chain
        self.fr = FieldRules()
        self.chained_df = None
        self.column_df = None
        self.file_format = self.settings['output']['file_format']
        self.appends = appends
        self.current_dir = os.path.dirname(sys.argv[0]) if not cli_path else cli_path

    def get_depend_on_file(self):
        """ find depends on other tables (data files)
            TODO: add support for multiple depends on files
        """
        dir_files = []
        if self.chains and self.table_id in self.chains:
            table_chain = self.chains[self.table_id]
            key = list(table_chain.keys())[0]
            src_file = os.path.join(self.current_dir, "{}.{}".format(
                table_chain[key]['table'], self.file_format))
            if not os.path.isfile(src_file):
                current_dir = os.getcwd()
                if current_dir in src_file:
                    src_file = os.path.join(current_dir, src_file)
                    if not os.path.isfile(src_file):
                        raise ValueError(f'{src_file} is not a file. Current dir {os.getcwd()}')
            dir_files.append(src_file)
        else:
            for item in self.schema:
                if item['name'] in self.chained:
                    chained_tables = [
                        table for table in self.chained[
                            item['name']] if table != self.table_id]
                    [dir_files.append(file_name) if file_name.startswith(table) else None
                     for table in chained_tables for file_name in os.listdir(
                        self.current_dir)]
                elif 'all' in self.chains and item['name'] in self.chains['all']:
                    table_chain = self.chains['all']
                    key = table_chain.keys()[0]
                    if table_chain[key]['table'] != self.table_id:
                        src_file = os.path.join(self.current_dir, "{}.{}".format(
                            table_chain[key]['table'], self.file_format))
                        if not os.path.isfile(src_file):
                            raise ValueError
                        dir_files.append(src_file)
        if dir_files:
            print("Depend on: {}".format(dir_files))
            return os.path.join(self.current_dir, dir_files[0])
        else:
            return []

    def schema_validation(self, schema):
        for item in schema:
            self.type_validation(item['type'])
        return schema

    @staticmethod
    def type_validation(type_name):
        if type_name not in supported_types:
            raise ValueError

    def _settings_validation(self, settings):
        raise NotImplementedError

    def resolve_dependencies(self):
        """ find dependencies on different ways and prepare self.chained_df"""
        depend_on_file = None
        if self.appends and self.table_id in self.appends:
            depend_on_file = [os.path.join(self.current_dir, "{}.{}".format(
                name, self.file_format)) for name in self.appends[self.table_id]]
        elif self.chained or self.chains:
            depend_on_file = self.get_depend_on_file()
        if depend_on_file:
            if isinstance(depend_on_file, str):
                self.chained_df = self._read_df_from_file(depend_on_file)
            elif isinstance(depend_on_file, Iterable):
                for file_name in depend_on_file:
                    if self.chained_df is None:
                        self.chained_df = self._read_df_from_file(file_name)
                    else:
                        self.chained_df = self.chained_df.merge(
                            how='outer', right=self._read_df_from_file(file_name))
        return self.chained_df

    def _read_df_from_file(self, depend_on_file):
        """ return correct pandas export method depend on expected file format """
        result = eval(f"read_{self.file_format}(\'{depend_on_file}\')")
        return result

    def init_timestamps_values(self):
        global current_time, last_time

        timezone = pytz.timezone(self.settings.get('timezone', 'UTC'))
        current_time = datetime.now(timezone).strftime("%Y-%m-%d-%H.%M.%S.%f")
        last_time = (datetime.now(timezone)
                     - timedelta(hours=48)).strftime("%Y-%m-%d-%H.%M.%S.%f")

    def get_data_frame(self):
        """ generate frame with data per table"""
        self.init_timestamps_values()
        logging.info("Start Data Generation")

        data_frame = DataFrame()
        current_obj = copy.deepcopy(self)
        num_cores = mp.cpu_count()
        pool = mp.Pool(num_cores)
        jobs = []

        for num, item in enumerate(self.schema):
            jobs.append(pool.apply_async(column_generation, (current_obj, item)))

            # wait for all jobs to finish
        for job in list(jobs):
            job.get()

        # clean up
        pool.close()

        num = 0
        while not q.empty():
            file_name = q.get()
            with open(file_name, 'r') as column_file:
                column = [line.split('\n')[0] for line in column_file.readlines()]
                data_frame.insert(num, column=file_name.split('tmp_')[1],
                                  value=Series(column))
                num += 1
            os.remove(file_name)
        return data_frame

    def get_field_rule(self, field_name):
        """ method to get field rule """
        field_rule = self.fr.rules.loc[self.fr.rules['field']
                                       == field_name].to_dict()
        if not field_rule.get('field'):
            field_rule = {}
        for elem in field_rule:
            field_rule[elem] = field_rule[elem][list(field_rule[elem].keys())[0]]
        return field_rule

    def get_values_from_chained_column(self, _field_name, _matches_k):
        """ get aliased column name and settings for chains:
        reversed and matches_k """
        src_column, revers = _field_name, False
        if self.chains and self.table_id in self.chains:
            for key in self.chains[self.table_id]:
                if key == _field_name:
                    src_column = self.chains[self.table_id][key]['alias']
                    if 'matches' in self.chains[self.table_id][key]:
                        _matches_k = self.chains[self.table_id][key]['matches']
                    if 'revers' in self.chains[self.table_id][key]:
                        revers = self.chains[self.table_id][key]['revers']
        return src_column, _matches_k, revers

    def get_dataframe_column(self, _src_column, _matches_k, _revers):
        """ get column for field from chained dataframe """
        df_column = self.chained_df[[_src_column]]
        splited_index = int(max(df_column.index) * _matches_k)
        if _revers:
            df_column = df_column.reindex(index=df_column.index[::-1])
        df_column = df_column.reindex(index=df_column.index[0:splited_index+1])
        df_column = df_column.to_dict()[_src_column].values()
        return df_column

    def get_column_from_chained(self, field_name, matches_k):
        alias = self.chains[self.table_id].get(field_name, {}).get('alias')
        field_name = alias or field_name
        df_column = None
        self.chained_df = self.resolve_dependencies()
        if self.chained_df is not None:
            src_column, matches_k, revers = self.get_values_from_chained_column(
                field_name, matches_k)
            if src_column in self.chained_df.columns:
                df_column = self.get_dataframe_column(src_column, matches_k, revers)
        return df_column

    def column_generator(self, field_name):
        """ create column with values """
        print("Generate column {}".format(field_name))
        column, unique = [], None
        table_fields_settings = self.settings.get('fields', {}).get(
            self.table_id, {})
        field_rule = self.get_field_rule(field_name)
        range_rows = self.settings.get('tables', {}).get(
            self.table_id, {}).get(
            'row_numbers') or self.settings['row_numbers']

        matches_k = self.settings['matches']
        if self.table_id in self.chains:
            df_column = self.get_column_from_chained(field_name, matches_k)
        else:
            df_column = None
        if table_fields_settings:
            if field_name in table_fields_settings:
                unique = table_fields_settings[field_name].get('unique') or range_rows
            elif (isinstance(
                    list(table_fields_settings.keys())[0], tuple)) and field_name \
                    in list(table_fields_settings.keys())[0]:
                unique = table_fields_settings[list(table_fields_settings.keys())[0]].get('unique') or range_rows
        if not unique:
            unique = range_rows
        print(unique)
        if df_column:
            [column.insert(0, elem) for elem in df_column]

        if len(column) < unique:
            unique = unique - len(column)
        else:
            column = column[:unique]
            unique = 0
        while unique:
            value = values_generator(field_rule)
            column.append(value)
            unique -= 1
        total_rows = range_rows - len(column)
        rel_size = total_rows / len(column)
        num_copy = int(rel_size)
        base_column = copy.deepcopy(column)
        for _ in range(num_copy):
            column += base_column
        float_adding = rel_size - num_copy
        column += base_column[:int(len(base_column) * float_adding)]

        return column


def column_generation(current_obj: DataGenerator,
                      _item: dict) -> None:
    file_name = 'tmp_{}'.format(_item['name'])
    with open(file_name, 'w+') as column_tmp:
        for line in current_obj.column_generator(_item['name']):
            column_tmp.write(str(line))
            column_tmp.write(str('\n'))
    q.put(file_name)




import string
import random  # noqa F401
import re  # noqa F401
from uuid import uuid1, uuid4  # noqa F401
from random import randint, random, choice, choices  # noqa F401


from mimesis import Generic, Datetime, Text, Address, Person


languages = ['zh', 'ru', 'en', 'es', 'ar', 'fr']


fake = Generic()
text = Text()
person = Person()
date = Datetime()
address = Address()


current_time = None
last_time = None


def random_char(_len):
    return ''.join([choice(string.ascii_letters) for x in range(_len)]).upper()


def values_generator(rule=None):
    """ values generator, that eval rules what was defined in 'generator' key  """

    value = None
    if rule and rule['generator']:
        value = eval(rule['generator'])
        if rule.get('len', '') and isinstance(rule['len'], int):
            value = value[0:rule['len']]
    return value



""" module that contain config object with default settings """
from copy import deepcopy


class Config(object):

    default_settings = {
        "row_numbers": 100,
        "matches": 0.6,
        "timezone": "UTC",
        "output": {"file_format": "json",
                   "config": {}},
    }

    cfg = None

    def __init__(self, settings):
        """
        :param settings:
        """
        self.settings = self.get_config(settings)

    def get_config(self, settings=None):
        """ get current config as is if settings is None or update and return config """
        if self.cfg is None:
            self.cfg = deepcopy(self.default_settings)
        if settings:
            self.cfg.update(settings)
            self.cfg = self._validate_settings(self.cfg)
        return self.cfg

    @staticmethod
    def _validate_settings(config):
        """ validate params in config  """

        if config.get("rowNumbers", 0) < 0 or config.get("matches", 0) < 0:
            raise ValueError("rowNumbers (int) and matches (float) must be positive")

        return config



""" classes to work with tables """

import os
import json
from typing import List, Dict, Text
from fakeme.generator import DataGenerator
from fakeme.output import get_writer


class SchemaExtractor(object):

    def __init__(self,
                 schema_path: Text,
                 folder: Text = None,
                 table_id: Text = "",
                 dump_schema: bool = True):
        self.schema_path = schema_path
        self.folder = folder or self.extract_folder()
        self.table_id = table_id
        self.dump_schema = dump_schema

    def extract_folder(self):
        splited = self.schema_path.split('/')
        folder, _ = splited[(len(splited) - 2):]
        if folder == 'schemas':
            folder = ''
        return folder

    def get_schema(self):
        """ get schema in BQ format """
        if self.schema_path.endswith('.json'):
            with open(self.schema_path, 'r') as schema_file:
                schema = json.load(schema_file)
        else:
            raise NotImplementedError('Supports only `.json` format')
        return schema


class TableRunner(object):
    """
    `schema_in_obj` arg example:

        [
            {"type": "STRING", "name": "name", "mode": "NULLABLE"},
            {"type": "STRING", "name": "id", "mode": "NULLABLE"}
        ]

    """
    def __init__(self,
                 dataset_id: Text,
                 table_id: Text,
                 settings: Dict,
                 schema_path: Text = None,
                 dump_schema: bool = False,
                 schema_in_obj: List = None):
        self.dataset_id = dataset_id
        self.table_id = table_id
        self.settings = settings
        if not schema_in_obj and not schema_path:
            raise ValueError("Impossible to create a table without a schema")
        if not schema_in_obj:
            self.schema = SchemaExtractor(schema_path=schema_path,
                                          folder=dataset_id,
                                          table_id=table_id,
                                          dump_schema=dump_schema).get_schema()
        else:
            if not isinstance(schema_in_obj, list):
                raise ValueError('`list_schema` must be list with fields dicts')
            self.schema = schema_in_obj

    def create_data(self,
                    file_path: Text,
                    chained: Dict = None,
                    alias_chain: Dict = None,
                    appends: Dict = None,
                    cli_path: Text = None,
                    remove_old: bool = True
                    ):
        """ generate data and save to the file """
        table_data = DataGenerator(
            schema=self.schema, settings=self.settings, chained=chained,
            table_id=self.table_id, alias_chain=alias_chain, appends=appends,
            cli_path=cli_path).get_data_frame()

        if table_data is None:
            raise Exception("Data Generator failed")
        else:
            target_file_path = self._prepare_path(file_path, remove_old)
            writer = get_writer(self.settings['output']['file_format'])
            writer(table_data, target_file_path, self.settings['output']['config'])

    @staticmethod
    def _prepare_path(file_path, remove_old):
        """ prepare folder and check target file """
        if os.path.isfile(file_path):
            print("Founded old file {}".format(file_path))
            if remove_old:
                os.remove(file_path)
                print("File {} was removed".format(file_path))
            else:
                raise Exception("Impossible to generate data into file {}. "
                                "File already exist. Please delete file or set "
                                "'remove_old'=True".format(file_path))
        else:
            if not os.path.isdir(os.path.abspath(os.path.dirname(file_path))):
                os.makedirs(os.path.dirname(file_path))
        return file_path


class MultiTableRunner(object):

    def __init__(self, tables_list: List, settings: Dict, prefix="."):
        """

        :param tables_list:
        :type tables_list: list
        :param settings:
        :type settings: dict
        :param prefix:
        :type prefix: str
        """
        self.tables = tables_list
        self.prefix = prefix
        self.settings = settings

    def _validate_path(self, path):
        """ check path exist, try to get correct path """
        path = os.path.expanduser(path)
        if not os.path.isfile(path):
            os.chdir(self.prefix)
            target_path = os.path.abspath(path)
        else:
            target_path = path
        if not os.path.isfile(target_path):
            raise ValueError("File with schema does not exist {}".format(path))
        return target_path

    def get_fields_and_schemas(self, dump_schema=False):
        """ return list of schemas and tables fields """
        schemas = {}
        fields = []
        for dataset_id, table_id, schema in self.tables:
            if isinstance(schema, str):
                schema = self._validate_path(schema)
                tg = TableRunner(dataset_id, table_id, schema_path=schema,
                                 settings=self.settings, dump_schema=dump_schema)
            else:
                tg = TableRunner(dataset_id, table_id, settings=self.settings,
                                 schema_in_obj=schema, dump_schema=dump_schema)
            schemas[table_id] = (tg.schema, tg)
        for table_id in schemas:
            fields.append((table_id, [item['name'] for item in schemas[table_id][0]]))
        return schemas, fields



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fakeme', 'fakeme.cli', 'fakeme.examples.space_ship_parts']

package_data = \
{'': ['*'], 'fakeme': ['examples/cli_usage/*', 'examples/cli_usage/schemas/*']}

install_requires = \
['mimesis>=4.0,<5.0', 'pandas>=1.0,<2.0', 'ply>=3.11,<4.0']

entry_points = \
{'console_scripts': ['fakeme = fakeme.cli:cli']}

setup_kwargs = {
    'name': 'fakeme',
    'version': '0.0.0.dev1',
    'description': 'Relative Dataframes Generator: generate tables with data, that depend each other',
    'long_description': None,
    'author': 'xnuinside',
    'author_email': 'xnuinside@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/xnuinside/fakeme',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from fakeme.core import RunGenerator
from fakeme.fields import FieldRules

# STEP 1: Define schemas
schema_one_parts_list = [{
  "type": "STRING",
  "name": "part_identification",
  "mode": "NULLABLE"
 },
 {
  "type": "STRING",
  "name": "ship_type",
  "mode": "NULLABLE"
 },
 {
  "type": "STRING",
  "name": "price",
  "mode": "NULLABLE"
 }
]


# STEP 2: Add rules for field generation if you want, if not - will be used default generation rules
FieldRules.user_rules.append(
    {"field": "count", "generator": "str(randint(100, 6000))", "len": ""})

# "generator" must contains code, that can be executed in Generator module with "eval" command
# to see that methods are exist in Generator that you can use - you can simple just check fakeme.generator module

# define more rule
ship_type = {"field": "ship_type", "generator": "'Ship ' + text.word()",  "len": ""}
FieldRules.user_rules.append(ship_type)

# create list of tables, each tuple - one table, values in indexes:
# 1st - dataset/database name
# 2nd - table name
# 3rd - table's schema

list_of_tables = [
    ('robot_factory', 'parts', schema_one_parts_list),
    ('robot_factory', 'warehouse', 'warehouse_schema.json')  # second schema we will read from the file
]

# STEP 3: define dependencies and generation rules
RunGenerator(tables_list=list_of_tables,
             dump_schema=True,
             params={'row_numbers': 15},  # how much rows we want to generate
             # rls stands for  relationship - defining relationship between tables,
             # that field depend on that
             rls={'warehouse': {'part_id': {'alias': 'part_identification',
                                            'matches': 1,
                                            'table': 'parts'}}
                  }).run()

# now just run `python space_ship_warehouse_tables.py`

# as result you will see 2 json files, that contains same data
# in part_identification (in parts.json) and part_id (in warehouse.json) fields
